#include "wTimeSolver.h"
#include "wBLRegion.h"
#include "wViscFluxes.h"
#include <Eigen/Dense>
#include <iostream>
#include <vector>
#include <cmath>

using namespace Eigen;
using namespace tbox;
using namespace dart;


TimeSolver::TimeSolver(double _CFL0, double _Minf, double Re, unsigned long _maxIt, double _tol, unsigned int _itFrozenJac)
{

  CFL0 = _CFL0;
  maxIt = _maxIt;
  tol = _tol;
  itFrozenJac0 = _itFrozenJac;
  Minf = std::max(_Minf, 0.1);

  SysMatrix = new ViscFluxes(Re);
  
} // end TimeSolver


TimeSolver::~TimeSolver()
{   
  delete SysMatrix;
  std::cout << "~TimeSolver()\n";
} // end ~TimeSolver

void TimeSolver::InitialCondition(size_t iPoint, BLRegion *bl)
{
  unsigned int nVar = bl->GetnVar();
  if (initSln){

    bl->U[iPoint * nVar + 0] = bl->U[(iPoint-1) * nVar + 0];
    bl->U[iPoint * nVar + 1] = bl->U[(iPoint-1) * nVar + 1];
    bl->U[iPoint * nVar + 2] = bl->U[(iPoint-1) * nVar + 2];
    bl->U[iPoint * nVar + 3] = bl->GetVtExt(iPoint);
    bl->U[iPoint * nVar + 4] = bl->U[(iPoint-1) * nVar + 4];

    bl->Ret[iPoint] = bl->Ret[iPoint - 1];
    bl->Cd[iPoint] = bl->Cd[iPoint - 1];
    bl->Cf[iPoint] = bl->Cf[iPoint - 1];
    bl->Hstar[iPoint] = bl->Hstar[iPoint - 1];
    bl->Hstar2[iPoint] = bl->Hstar2[iPoint - 1];
    bl->Hk[iPoint] = bl->Hk[iPoint - 1];
    bl->Cteq[iPoint] = bl->Cteq[iPoint - 1];
    bl->Us[iPoint] = bl->Us[iPoint - 1];
    bl->Delta[iPoint] = bl->Delta[iPoint - 1];
    bl->DeltaStar[iPoint] = bl->DeltaStar[iPoint - 1];
  }
  if (bl->Regime[iPoint] == 1 && bl->U[iPoint * nVar + 4] == 0){
      bl->U[iPoint * nVar + 4] = 0.03;
  } // End if
  else if (bl->Regime[iPoint] == 0){
    bl->U[iPoint*nVar+4] = 0;
  } // End else if
} // End InitialCondition

int TimeSolver::Integration(size_t iPoint, BLRegion *bl)
{

  // Save initial condition

  unsigned int nVar = bl->GetnVar();

  std::vector<double> temp_U;
  for (size_t i=0; i<nVar; ++i){
    temp_U.push_back(bl->U[iPoint*nVar+i]);
  } // End for

  // Initialize time integration variables

  double dx = bl->GetLocMarkers(iPoint) - bl->GetLocMarkers(iPoint-1);

  double CFL = CFL0;
  unsigned int itFrozenJac = itFrozenJac0;

  double timeStep = SetTimeStep(CFL, dx, bl->GetVtExt(iPoint));

  Matrix<double, 5, 5> IMatrix;
  IMatrix = Matrix<double, 5, 5>::Identity() / timeStep;

  Vector<double, 5> SysRes = SysMatrix->ComputeFluxes(iPoint, bl);
  double normSysRes0 = SysRes.norm();
  double normSysRes = normSysRes0;

  Matrix<double, 5, 5> JacMatrix;
  Vector<double, 5> slnIncr;
  
  nErrors = 0; // Number of errors encountered
  unsigned int innerIters = 0; // Inner (non-linear) iterations

  while (innerIters < maxIt){

    if(innerIters % itFrozenJac == 0)
    {
      if (nErrors > 4)
      {
        return -1; // Convergence failed
      }
      CFL = ControlIntegration(iPoint, bl, temp_U, CFL);
      itFrozenJac = itFrozenJac0;
      JacMatrix = SysMatrix->ComputeJacobian(iPoint, bl, SysRes, 1e-8);
    }

    slnIncr = (JacMatrix + IMatrix).partialPivLu().solve(-SysRes);

    for(size_t k=0; k<nVar; ++k)
    {
      bl->U[iPoint*nVar+k] += slnIncr(k);
    }
    if (itFrozenJac0 == 1)
    {
      bl->U[iPoint*nVar + 0] = std::max(bl->U[iPoint*nVar + 0], 1e-7);
      bl->U[iPoint*nVar + 1] = std::max(bl->U[iPoint*nVar + 1], 1.0005);
    }


    SysRes = SysMatrix->ComputeFluxes(iPoint, bl);
    normSysRes = (SysRes + slnIncr/timeStep).norm();
    if (bl->name == "wake" && iPoint == 20)
    {
      std::cout << normSysRes/normSysRes0 << std::scientific;
      bl->PrintSolution(iPoint);
    }

    if (normSysRes <= tol * normSysRes0)
    {
      return 0; /* Successfull exit */
    }

    // CFL adaptation.

    CFL = CFL0* pow(normSysRes0/normSysRes, 0.7);
    CFL = std::max(CFL, 1.0);
    timeStep = SetTimeStep(CFL, dx, bl->GetVtExt(iPoint));
    IMatrix = Matrix<double, 5, 5>::Identity() / timeStep;

    innerIters++;
  } // End while (innerIters < maxIt)

  for (size_t k=0; k<nVar; ++k)
  {
    if (std::isnan(bl->U[iPoint*nVar + k]))
    {
      ResetSolution(iPoint, bl, temp_U);
      return innerIters; // New CFL.
    }
  }
  if (normSysRes <= 1e-4 * normSysRes0)
  {
    return 0;
  }
  else
  {
    return innerIters;
  }
   

  ResetSolution(iPoint, bl, temp_U);
  return innerIters;
} // End Integration

double TimeSolver::SetTimeStep(double CFL, double dx, double invVel)
{
  /* Speed of sound. */
  double gradU2 = (invVel*invVel);
  double soundSpeed = ComputeSoundSpeed(gradU2);

  /* Velocity of the fastest travelling wave */
  double advVel = soundSpeed + invVel;

  /* Time step computation */
  return CFL*dx/advVel;
}

double TimeSolver::ComputeSoundSpeed(double gradU2)
{
  return sqrt(1 / (Minf * Minf) + 0.5 * (gamma - 1) * (1 - gradU2));
}

double TimeSolver::ControlIntegration(size_t iPoint, BLRegion *bl, std::vector<double> temp_U, double CFLIn){

  /* Check if there are NaN in the solution */
  unsigned int nVar = bl->GetnVar();
  for (size_t k=0; k<nVar; ++k)
  {
    if (std::isnan(bl->U[iPoint*nVar + k]))
    {
      //std::cout << "Point " << iPoint << " : Solution reset; NaN identified at position " << k << std::endl;
      ResetSolution(iPoint, bl, temp_U);
      itFrozenJac0 = 1; // Impose continuous Jacobian evaluation.
      nErrors+=1;
      return 0.3; // New CFL.
    }
  }
  return CFLIn; // Keep CFL.
}

void TimeSolver::ResetSolution(size_t iPoint, BLRegion *bl, std::vector<double> temp_U){

  //std::cout << "Reseting point" << iPoint << std::endl;
  unsigned int nVar = bl->GetnVar();

  /* Reset solution */

  for(size_t k=0; k<nVar; ++k)
  {
    if (std::isnan(bl->U[iPoint*nVar+k]))
    {
      for (size_t j = 0; k<bl->GetnVar(); ++j)
      {
        if(!std::isnan(temp_U[j]))
        {
        bl->U[iPoint*bl->GetnVar()+j] = temp_U[j];
        }
        else
        {
           bl->U[iPoint*nVar+j] = bl->U[(iPoint-1)*nVar+j];
        }
      break;
    } // End if
  } // End for

  /* Reset closures */

  if (bl->Regime[iPoint] == 0){
    std::vector<double> lamParam = bl->closSolver->LaminarClosures(bl->U[iPoint*nVar+0], bl->U[iPoint*nVar+1], bl->Ret[iPoint], bl->GetMe(iPoint), bl->name);
    bl->Hstar[iPoint] = lamParam[0];
    bl->Hstar2[iPoint] = lamParam[1];
    bl->Hk[iPoint] = lamParam[2];
    bl->Cf[iPoint] = lamParam[3];
    bl->Cd[iPoint] = lamParam[4];
    bl->Delta[iPoint] = lamParam[5];
    bl->Cteq[iPoint] = lamParam[6];
    bl->Us[iPoint] = lamParam[7];
  } // End if
  else if (bl->Regime[iPoint] == 1){
    std::vector<double> turbParam = bl->closSolver->TurbulentClosures(bl->U[iPoint*nVar+0], bl->U[iPoint*nVar+1], bl->U[iPoint*nVar+4], bl->Ret[iPoint], bl->GetMe(iPoint), bl->name);
    bl->Hstar[iPoint] = turbParam[0];
    bl->Hstar2[iPoint] = turbParam[1];
    bl->Hk[iPoint] = turbParam[2];
    bl->Cf[iPoint] = turbParam[3];
    bl->Cd[iPoint] = turbParam[4];
    bl->Delta[iPoint] = turbParam[5];
    bl->Cteq[iPoint] = turbParam[6];
    bl->Us[iPoint] = turbParam[7];
  } // End else if

  } // End for
}


double TimeSolver::GetCFL0(){return CFL0;}
void TimeSolver::SetCFL0(double _CFL0){CFL0 = _CFL0;}
unsigned long TimeSolver::GetmaxIt(){return maxIt;}
void TimeSolver::SetmaxIt(unsigned long _maxIt){maxIt = _maxIt;}
unsigned int TimeSolver::GetitFrozenJac(){return itFrozenJac0;}
void TimeSolver::SetitFrozenJac(unsigned int _itFrozenJac){itFrozenJac0 = _itFrozenJac;}

void TimeSolver::SetinitSln(bool _initSln){initSln = _initSln;}