#include "wViscSolver.h"
#include "wBLRegion.h"
#include "wTimeSolver.h"
#include <iostream>
#include <vector>


using namespace tbox;
using namespace dart;


ViscSolver::ViscSolver(double _Re, double _Minf, double _CFL0)
{

  /* Freestream parameters */
  Re = _Re;
  Minf = _Minf;

  /* User input numerical parameters */
  CFL0 = _CFL0;

  std::cout << "--- Viscous solver setup ---\n"
              << "Reynolds number: " << Re << " \n"
              << "Freestream Mach number: " << Minf << " \n"
              << "Initial CFL number: " << CFL0 << " \n"
              << std::endl;

  /* Initialzing boundary layer */

  for (size_t i=0; i<3; ++i)
  {
    BLRegion *region = new BLRegion();
    bl.push_back(region);
  } // end for

  tSolver = new TimeSolver(CFL0, Minf, Re);

  bl[0]->name = "upper";
  bl[1]->name = "lower";
  bl[2]->name = "wake";

  nbElmUpper = 0;
  stagPtMvmt = false;

  std::cout << "Initializing boundary layer" << std::endl;

} // end ViscSolver


ViscSolver::~ViscSolver()
{
  for (size_t i=0; i<bl.size(); ++i){
    delete bl[i];
  } // end for

  delete tSolver;

  std::cout << "~ViscSolver()\n";


} // end ~ViscSolver

void ViscSolver::InitMeshes(size_t region, std::vector<double> locM, std::vector<double> globM)
{
  bl[region]->SetMesh(locM, globM);
}

void ViscSolver::SendInvBC(size_t region, std::vector<double> _Ue,
                                         std::vector<double> _Me,
                                         std::vector<double> _Rhoe)
{
  bl[region]->SetInvBC(_Ue, _Me, _Rhoe);
}

void ViscSolver::SendExtVariables(size_t region, std::vector<double> _xxExt, std::vector<double> _deltaStarExt)
{
  bl[region]->SetExtVariables(_xxExt, _deltaStarExt);
}


int ViscSolver::Run(unsigned int couplIter){

  /* Prepare time integration */

  if ((couplIter == 0) || (nbElmUpper != bl[0]->GetnMarkers()) || (stagPtMvmt)){
    tSolver->SetCFL0(0.5);
    tSolver->SetitFrozenJac(1);
    tSolver->SetinitSln(true);
    if (couplIter != 0 && (nbElmUpper != bl[0]->GetnMarkers())){
      stagPtMvmt = true;
      std::cout << "Stagnation point moved" << std::endl;
    }
    else{
      stagPtMvmt = false;
    }
    std::cout << "Restart solution" << std::endl;
  }

  else{
    tSolver->SetCFL0(0.5);
    tSolver->SetitFrozenJac(1);
    tSolver->SetinitSln(false);
    stagPtMvmt = false;
    std::cout << "Updating solution" << std::endl;
  }

  nbElmUpper = bl[0]->GetnMarkers();

  /* Set boundary condition */

  bl[0]->SetStagBC(Re);
  bl[1]->SetStagBC(Re);

  /* Loop over the regions */

  bool lockTrans;
  int solverExitCode = 0;
  int pointExitCode;

  for (size_t iRegion = 0; iRegion < bl.size(); ++iRegion){

    /* Reset regime */
    if (iRegion<2){
      for (size_t k=0; k<bl[iRegion]->GetnMarkers(); k++){
        bl[iRegion]->Regime[k]=0;
      }
    }
    else{
      for (size_t k=0; k<bl[iRegion]->GetnMarkers(); k++){
        bl[iRegion]->Regime[k]=1;
      }
    }
    lockTrans = false;

    std::cout << "- " << bl[iRegion]->name << " region";

    if (iRegion == 2){
      /* Impose wake boundary condition */
      std::vector<double> UpperCond;
      std::vector<double> LowerCond;
      UpperCond = std::vector<double>(bl[0]->U.end() - bl[0]->GetnVar(), bl[0]->U.end()); /* Base std::vector constructor to slice */
      LowerCond = std::vector<double>(bl[1]->U.end() - bl[1]->GetnVar(), bl[1]->U.end()); /* Base std::vector constructor to slice */
      bl[iRegion]->SetWakeBC(Re, UpperCond, LowerCond);
      std::cout << "\n" << std::endl;

      lockTrans = true;
    }

    /* Loop over points */ 

    for (size_t iPoint = 1; iPoint < bl[iRegion]->GetnMarkers(); ++iPoint){

      /* Initialize solution at point */

      tSolver->InitialCondition(iPoint, bl[iRegion]);

      /* Time integration */

      pointExitCode = tSolver->Integration(iPoint, bl[iRegion]);
      if (pointExitCode!=0)
      {
        std::cout << "Point " << iPoint << ": Convergence terminated with exitcode " << pointExitCode << std::endl;
        solverExitCode = -1; // Convergence failed
      }

      /* Transition */

      if (!lockTrans)
      {
        CheckWaves(iPoint, bl[iRegion]);

        // Free transition.
        if (bl[iRegion]->U[iPoint*bl[iRegion]->GetnVar() + 2] >= bl[iRegion]->GetnCrit()){
          AverageTransition(iPoint, bl[iRegion], 0);
          std::cout << ", free transition x/c = " << bl[iRegion]->xtr << ". Marker " << bl[iRegion]->transMarker << std::endl;
          lockTrans = true;
        } // End if
      } // End if (!lockTrans).

    } // End for iPoints
  } // End for iRegion

  ComputeDrag();

  for (size_t iRegion=0; iRegion<bl.size(); ++iRegion)
  {
    bl[iRegion]->ComputeBlwVel();
  } // End for iRegion

  return solverExitCode; // exit code
  }

void ViscSolver::CheckWaves(size_t iPoint, BLRegion *bl)
{
  /* Determine if the amplification waves start growing or not. */

  double logRet_crit = 2.492*pow(1/(bl->Hk[iPoint]-1), 0.43)
  + 0.7*(tanh(14*(1/(bl->Hk[iPoint]-1))-9.24)+1.0);

  if (bl->Ret[iPoint] > 0)
  { // Avoid log of negative number.
    if (log10(bl->Ret[iPoint]) <= logRet_crit - 0.08){
      bl->U[iPoint*bl->GetnVar() + 2] = 0; // Set N to 0 at that point if waves do not grow.
    }
  }
}

void ViscSolver::AverageTransition(size_t iPoint, BLRegion *bl, int forced)
{
  /* Averages solution on transition marker */
  
  unsigned int nVar = bl->GetnVar();

  // Save laminar solution.
  std::vector<double> lamSol;
  for (size_t k=0; k<nVar; ++k)
  {
    lamSol.push_back(bl->U[iPoint*nVar + k]);
  }

  // Set up turbulent portion boundary condition.
  bl->transMarker = iPoint; /* Save transition marker */

  /* Loop over remaining points in the region */
  for (size_t k=0; k<bl->GetnMarkers() - iPoint; ++k)
  {
    bl->Regime[iPoint+k] = 1; // Set Turbulent regime. 
  }

  /* Compute transition location */
  bl->xtr = (bl->GetnCrit()-bl->U[(iPoint-1)*nVar + 2])*((bl->GetGlobMarkers(iPoint)-bl->GetGlobMarkers(iPoint-1))/(bl->U[iPoint*nVar + 2]-bl->U[(iPoint-1)*nVar+2]))+bl->GetGlobMarkers(iPoint-1);
  
  /*  Percentage of the subsection corresponding to a turbulent flow */
  double avTurb = (bl->GetGlobMarkers(iPoint)-bl->xtr)/(bl->GetGlobMarkers(iPoint)-bl->GetGlobMarkers(iPoint-1));
  double avLam = 1-avTurb;

  /* Impose boundary condition */
  std::vector<double> turbParam1 = bl->closSolver->TurbulentClosures(bl->U[(iPoint-1)*nVar + 0], bl->U[(iPoint-1)*nVar + 1], 0.03, bl->Ret[iPoint-1], bl->GetMe(iPoint-1), bl->name);
  bl->Cteq[iPoint-1] = turbParam1[6];
  bl->U[(iPoint-1)*nVar + 4] = 0.7 * bl->Cteq[iPoint-1];

  /* Avoid starting with ill conditioned IC for turbulent computation. (Regime was switched line 183) */
  /* These initial guess values do not influence the converged solution */
  bl->U[iPoint*nVar + 4] = bl->U[(iPoint-1)*nVar + 4]; // IC of transition point = turbulent BC (imposed @ previous point).
  bl->U[iPoint*nVar + 1] = 1.515; // Because we expect a significant drop of the shape factor.

  /* Solve point in turbulent configuration */
  tSolver->Integration(iPoint, bl);

  /* Average both solutions (Now solution @ iPoint is the turbulent solution) */
  bl->U[iPoint*nVar + 0] = avLam * lamSol[0] + avTurb * bl->U[(iPoint)*nVar + 0]; /* Theta */
  bl->U[iPoint*nVar + 1] = avLam * lamSol[1] + avTurb * bl->U[(iPoint)*nVar + 1]; /* H */
  bl->U[iPoint*nVar + 2] = 9;                                                     /* N */
  bl->U[iPoint*nVar + 3] = avLam * lamSol[3] + avTurb * bl->U[(iPoint)*nVar + 3]; /* ue */
  bl->U[iPoint*nVar + 4] = avTurb * bl->U[(iPoint)*nVar + 4];                     /* Ct */

  /* Recompute closures. (The turbulent BC @ iPoint - 1 does not influence laminar closure relations). */
  std::vector<double> lamParam = bl->closSolver->LaminarClosures(bl->U[(iPoint-1)*nVar + 0], bl->U[(iPoint-1)*nVar + 1], bl->Ret[iPoint-1], bl->GetMe(iPoint-1), bl->name);
  bl->Hstar[iPoint-1] = lamParam[0];
  bl->Hstar2[iPoint-1] = lamParam[1];
  bl->Hk[iPoint-1] = lamParam[2];
  bl->Cf[iPoint-1] = lamParam[3];
  bl->Cd[iPoint-1] = lamParam[4];
  bl->Delta[iPoint-1] = lamParam[5];
  bl->Cteq[iPoint-1] = lamParam[6];
  bl->Us[iPoint-1] = lamParam[7];

  std::vector<double> turbParam = bl->closSolver->TurbulentClosures(bl->U[(iPoint)*nVar + 0], bl->U[(iPoint)*nVar + 1], bl->U[(iPoint)*nVar + 4], bl->Ret[iPoint], bl->GetMe(iPoint), bl->name);
  bl->Hstar[iPoint] = turbParam[0];
  bl->Hstar2[iPoint] = turbParam[1];
  bl->Hk[iPoint] = turbParam[2];
  bl->Cf[iPoint] = turbParam[3];
  bl->Cd[iPoint] = turbParam[4];
  bl->Delta[iPoint] = turbParam[5];
  bl->Cteq[iPoint] = turbParam[6];
  bl->Us[iPoint] = turbParam[7];
}

std::vector<double> ViscSolver::ExtractBlowingVel(size_t iRegion)
{
  return bl[iRegion]->Blowing;
}
std::vector<double> ViscSolver::ExtractDeltaStar(size_t iRegion)
{
  return bl[iRegion]->DeltaStar;
}
std::vector<double> ViscSolver::ExtractXx(size_t iRegion)
{
  std::vector<double> LocMarkersOut;
  for (size_t k=0; k<bl[iRegion]->GetnMarkers(); ++k){
    LocMarkersOut.push_back(bl[iRegion]->GetLocMarkers(k));
  }
  return LocMarkersOut;
}


void ViscSolver::ComputeDrag()
{

  unsigned int nVar = bl[0]->GetnVar();
  unsigned int lastWkPt = (bl[2]->GetnMarkers() - 1)*nVar;

  /* Normalize friction and dissipation coefficients. */

  std::vector<double> frictioncoeff_upper;
  for (size_t k=0; k<bl[0]->GetnMarkers(); ++k){
    frictioncoeff_upper.push_back(bl[0]->U[k*nVar + 3]*bl[0]->U[k*nVar + 3] * bl[0]->Cf[k]);
  }
  std::vector<double> frictioncoeff_lower;
  for (size_t k=0; k<bl[1]->GetnMarkers(); ++k){
    frictioncoeff_lower.push_back(bl[1]->U[k*nVar + 3]*bl[1]->U[k*nVar + 3] * bl[1]->Cf[k]);
  }

  /* Compute friction drag coefficient. */

  double Cdf_upper = 0.0;
  for (size_t k=1; k<bl[0]->GetnMarkers(); ++k){
    Cdf_upper += (frictioncoeff_upper[k]+frictioncoeff_upper[k-1])*(bl[0]->GetGlobMarkers(k) - bl[0]->GetGlobMarkers(k-1))/2;
  }
  double Cdf_lower = 0.0;
  for (size_t k=1; k<bl[1]->GetnMarkers(); ++k){
    Cdf_lower += (frictioncoeff_lower[k]+frictioncoeff_lower[k-1])*(bl[1]->GetGlobMarkers(k) - bl[1]->GetGlobMarkers(k-1))/2;

  }

  Cdf = Cdf_upper + Cdf_lower; // No friction in the wake
  
  /* Compute total drag coefficient. */

  Cdt = 2 * bl[2]->U[lastWkPt+0] * pow(bl[2]->U[lastWkPt+3],(bl[2]->U[lastWkPt+1]+5)/2);
  
  /* Compute pressure drag coefficient. */

  Cdp = Cdt - Cdf;
}